"""
This module defines an detector for minimal transition times between states
(e.g. value combinations of stated paths).
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os
import logging
import sys
import time

from aminer.AminerConfig import DEBUG_LOG_NAME, build_persistence_file_name, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX,\
    KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class MinimalTransitionTimeDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """This class creates events when minimal transition times between states (e.g. value combinations of stated paths) are undercut."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, anomaly_event_handlers, target_path_list=None, id_path_list=None, ignore_list=None,
                 allow_missing_id=False, num_log_lines_solidify_matrix=100, time_output_threshold=0, anomaly_threshold=0.05,
                 persistence_id='Default', learn_mode=False, output_logline=True, stop_learning_time=None,
                 stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that values are analyzed by their combined
               occurrences. When no paths are specified, the events given by the full path list are analyzed.
        @param id_path_list the list of paths where id values can be stored in all relevant log event types.
        @param ignore_list list of paths that are not considered for analysis, i.e., events that contain one of these paths are
               omitted. The default value is [] as None is not iterable.
        @param allow_missing_id specifies whether log atoms without id path should be omitted (only if id path is set).
               does not refer to an existing parsed data object.
        @param num_log_lines_solidify_matrix number of processed log lines after which the matrix is solidified.
               This process is periodically repeated.
        @param time_output_threshold threshold for the tested minimal transition time which has to be exceeded to be tested.
        @param anomaly_threshold threshold for the confidence which must be exceeded to raise an anomaly.
        @param persistence_id name of persistence file.
        @param learn_mode specifies whether newly observed sequences should be added to the learned model
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            mutable_default_args=["target_path_list", "id_path_list", "ignore_list"], aminer_config=aminer_config,
            anomaly_event_handlers=anomaly_event_handlers, target_path_list=target_path_list, id_path_list=id_path_list,
            ignore_list=ignore_list, allow_missing_id=allow_missing_id, num_log_lines_solidify_matrix=num_log_lines_solidify_matrix,
            time_output_threshold=time_output_threshold, anomaly_threshold=anomaly_threshold, persistence_id=persistence_id,
            learn_mode=learn_mode, output_logline=output_logline, stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )

        # Test if both target_path_list and id_path_list are not empty
        if [] in (self.target_path_list, self.id_path_list):
            msg = 'Both target_path_list and id_path_list must not be empty.'
            logging.getLogger(DEBUG_LOG_NAME).warning(msg)
            print('WARNING: ' + msg, file=sys.stderr)

        # Initialization auxiliary variables
        self.time_matrix = {}
        self.last_value = {}
        self.last_time = {}
        self.log_total = 0

        # Load persistence
        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        PersistenceUtil.add_persistable_component(self)
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        if persistence_data is not None:
            return_matrix = persistence_data[0]
            keys_1 = [tuple(key) for key in persistence_data[1]]
            keys_2 = [[tuple(key) for key in persistence_data[2][i]] for i in range(len(persistence_data[2]))]
            self.time_matrix = {keys_1[i]: {keys_2[i][j]: return_matrix[i][j] for j in range(len(keys_2[i]))} for i in range(len(keys_1))}

    def receive_atom(self, log_atom):
        """Receive a log atom from a source and analyzes minimal times between transitions."""
        parser_match = log_atom.parser_match

        # Do not analyze the log line if target_path_list or id_path_list is empty
        if [] in (self.target_path_list, self.id_path_list):
            return False

        # Skip paths from ignore list.
        if any(ignore_path in parser_match.get_match_dictionary().keys() for ignore_path in self.ignore_list):
            return False

        # Skip line if atom_time is not defined.
        if log_atom.atom_time is None:
            return False

        # Increase the count by one and check if the matrix should be solidified.
        self.log_total += 1
        if self.log_total % self.num_log_lines_solidify_matrix == 0:
            self.solidify_matrix()

        # Use target_path_list to differentiate sequences by their id.
        event_value = ()
        for path in self.target_path_list:
            match = log_atom.parser_match.get_match_dictionary().get(path)
            if match is None:
                if self.allow_missing_id is True:
                    # Insert placeholder for path that is not available
                    event_value += ('',)
                else:
                    # Omit log atom if one of the id paths is not found.
                    return False
            else:
                if isinstance(match.match_object, bytes):
                    event_value += (match.match_object.decode(AminerConfig.ENCODING),)
                else:
                    event_value += (match.match_object,)

        # Get current index from combination of values of paths of id_path_list
        id_tuple = ()
        for id_path in self.id_path_list:
            id_match = log_atom.parser_match.get_match_dictionary().get(id_path)
            if id_match is None:
                if self.allow_missing_id is True:
                    # Insert placeholder for id_path that is not available
                    id_tuple += ('',)
                else:
                    # Omit log atom if one of the id paths is not found.
                    return False
            else:
                if isinstance(id_match.match_object, bytes):
                    id_tuple += (id_match.match_object.decode(AminerConfig.ENCODING),)
                else:
                    id_tuple += (id_match.match_object,)

        # Check if id_tuple has already appeared.
        if id_tuple not in self.last_value:
            # Initialize the last value and time
            self.last_value[id_tuple] = event_value
            self.last_time[id_tuple] = log_atom.atom_time
        else:
            # Check if the event_value changed or if the times are not strictly ascending and skip the line in that cases.
            if self.last_value[id_tuple] == event_value:
                self.last_time[id_tuple] = log_atom.atom_time
                return True
            if log_atom.atom_time - self.last_time[id_tuple] < 0:
                additional_information = {'AffectedLogAtomValues': [list(self.last_value[id_tuple]), list(event_value)],
                                          'AffectedIdValues': list(id_tuple), 'PreviousTime': self.last_time[id_tuple],
                                          'NewTime': log_atom.atom_time}
                self.print(f'Anomaly in log line order: {list(self.last_value[id_tuple])} - {list(event_value)} ({list(id_tuple)}): '
                           f'{self.last_time[id_tuple]} - {log_atom.atom_time}', log_atom, self.target_path_list, confidence=1,
                           additional_information=additional_information)
                return True

            # Check in which order the event_values appear in the time matrix
            event_value_1 = None
            event_value_2 = None
            if event_value in self.time_matrix and self.last_value[id_tuple] in self.time_matrix[event_value]:
                event_value_1 = event_value
                event_value_2 = self.last_value[id_tuple]
            elif self.last_value[id_tuple] in self.time_matrix and event_value in self.time_matrix[self.last_value[id_tuple]]:
                event_value_1 = self.last_value[id_tuple]
                event_value_2 = event_value

            if event_value_1 is None:
                # Initialize the entry in the time matrix
                if event_value not in self.time_matrix:
                    self.time_matrix[event_value] = {}

                additional_information = {'AffectedLogAtomValues': [list(self.last_value[id_tuple]), list(event_value)],
                                          'AffectedIdValues': list(id_tuple),
                                          'NewMinimalTime': log_atom.atom_time - self.last_time[id_tuple]}
                message = f'First Appearance: {list(self.last_value[id_tuple])} - {list(event_value)} ({list(id_tuple)}),' \
                          f' {log_atom.atom_time - self.last_time[id_tuple]}'
                self.print(message, log_atom, self.target_path_list, additional_information=additional_information)
                if self.learn_mode:
                    self.time_matrix[event_value][self.last_value[id_tuple]] = log_atom.atom_time - self.last_time[id_tuple]
                    if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                        self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time
            else:
                # Check and update if the time was under cut
                if self.time_matrix[event_value_1][event_value_2] > log_atom.atom_time - self.last_time[id_tuple] and\
                        self.time_matrix[event_value_1][event_value_2] > self.time_output_threshold:
                    if 1 - (log_atom.atom_time - self.last_time[id_tuple]) / self.time_matrix[event_value_1][event_value_2] >\
                            self.anomaly_threshold:
                        additional_information = {'AffectedLogAtomValues': [list(self.last_value[id_tuple]), list(event_value)],
                                                  'AffectedIdValues': list(id_tuple),
                                                  'PreviousMinimalTime': self.time_matrix[event_value_1][event_value_2],
                                                  'NewMinimalTime': log_atom.atom_time - self.last_time[id_tuple]}
                        message = f'Undercut transition time: {list(self.last_value[id_tuple])} - {list(event_value)} ({list(id_tuple)}),' \
                                  f' {self.time_matrix[event_value_1][event_value_2]} -> {log_atom.atom_time - self.last_time[id_tuple]}'
                        confidence = 1 - (log_atom.atom_time - self.last_time[id_tuple]) / self.time_matrix[event_value_1][event_value_2]
                        self.print(
                            message, log_atom, self.target_path_list, confidence=confidence, additional_information=additional_information)

                    if self.learn_mode:
                        self.time_matrix[event_value_1][event_value_2] = log_atom.atom_time - self.last_time[id_tuple]
                        if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                            self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time

            # Update the last_value and time
            self.last_value[id_tuple] = event_value
            self.last_time[id_tuple] = log_atom.atom_time

        return True

    def solidify_matrix(self):
        """Solidify minimal time matrix with the trianlge inequality."""
        # Initialize list old_pairs with all transitions and a list of all values
        # The list of old_pairs includes the minimal times which can be used to reduce the minimal ransition times of other transitions
        values = list(self.time_matrix.keys())
        for key1 in self.time_matrix:
            values += [key for key in self.time_matrix[key1] if key not in values]
        old_pairs = [[key1, key2] for key1 in self.time_matrix for key2 in self.time_matrix[key1]]

        # Check the triangle inequality as long as values are corrected
        while len(old_pairs) > 0:
            new_pairs = []
            for old_pair in old_pairs:
                # Check triangle inequality value - old_pair[0] - old_pair[1] > value - old_pair[1] and
                # old_pair[0] - old_pair[1] - value > value - old_pair[0]
                for value in values:
                    if value in (old_pair[0], old_pair[1]):
                        continue

                    # Check value - old_pair[0] - old_pair[1] > value - old_pair[1]
                    if (old_pair[0] in self.time_matrix and value in self.time_matrix[old_pair[0]]) or (
                           value in self.time_matrix and old_pair[0] in self.time_matrix[value]):

                        if old_pair[0] in self.time_matrix and value in self.time_matrix[old_pair[0]]:
                            key_1_1 = old_pair[0]
                            key_1_2 = value
                        else:
                            key_1_1 = value
                            key_1_2 = old_pair[0]

                        if old_pair[1] in self.time_matrix and value in self.time_matrix[old_pair[1]]:
                            key_2_1 = old_pair[1]
                            key_2_2 = value
                        else:
                            key_2_1 = value
                            key_2_2 = old_pair[1]

                        if key_2_1 not in self.time_matrix:
                            self.time_matrix[key_2_1] = {}
                        if (key_2_2 not in self.time_matrix[key_2_1] or self.time_matrix[key_1_1][key_1_2] +
                                self.time_matrix[old_pair[0]][old_pair[1]] < self.time_matrix[key_2_1][key_2_2]):
                            self.time_matrix[key_2_1][key_2_2] = self.time_matrix[key_1_1][key_1_2] +\
                                    self.time_matrix[old_pair[0]][old_pair[1]]
                            if [key_2_1, key_2_2] not in new_pairs:
                                new_pairs += [[key_2_1, key_2_2]]

                    # Check old_pair[0] - old_pair[1] - value > value - old_pair[0]
                    if (old_pair[1] in self.time_matrix and value in self.time_matrix[old_pair[1]]) or (
                           value in self.time_matrix and old_pair[1] in self.time_matrix[value]):

                        if old_pair[1] in self.time_matrix and value in self.time_matrix[old_pair[1]]:
                            key_1_1 = old_pair[1]
                            key_1_2 = value
                        else:
                            key_1_1 = value
                            key_1_2 = old_pair[1]

                        if old_pair[0] in self.time_matrix and value in self.time_matrix[old_pair[0]]:
                            key_2_1 = old_pair[0]
                            key_2_2 = value
                        else:
                            key_2_1 = value
                            key_2_2 = old_pair[0]

                        if key_2_1 not in self.time_matrix:
                            self.time_matrix[key_2_1] = {}
                        if (key_2_2 not in self.time_matrix[key_2_1] or self.time_matrix[key_1_1][key_1_2] +
                                self.time_matrix[old_pair[0]][old_pair[1]] < self.time_matrix[key_2_1][key_2_2]):
                            self.time_matrix[key_2_1][key_2_2] = self.time_matrix[key_1_1][key_1_2] +\
                                    self.time_matrix[old_pair[0]][old_pair[1]]
                            if [key_2_1, key_2_2] not in new_pairs:
                                new_pairs += [[key_2_1, key_2_2]]

            old_pairs = new_pairs

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        persist_data = []
        keys_1 = list(self.time_matrix.keys())
        keys_2 = [list(self.time_matrix[key].keys()) for key in keys_1]
        return_matrix = [[self.time_matrix[keys_1[i]][keys_2[i][j]] for j in range(len(keys_2[i]))] for i in range(len(keys_1))]
        persist_data.append(return_matrix)
        persist_data.append(keys_1)
        persist_data.append(keys_2)
        PersistenceUtil.store_json(self.persistence_file_name, persist_data)
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.constraint_list:
            self.constraint_list.append(event_data)
        return f'Allowlisted path {event_data}.'

    def print_persistence_event(self, event_type, event_data):
        """
        Prints the persistence of component_name. Event_data specifies what information is output.
        @return a message with information about the persistence.
        @throws Exception when the output for the event_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        # Query if event_data has one of the stated formats
        if not (isinstance(event_data, list) and len(event_data) <= 2 and (
                    (len(event_data) == 2 and isinstance(event_data[0], list) and isinstance(event_data[1], list) and
                     len(event_data[0]) == len(self.target_path_list) and len(event_data[1]) == len(self.target_path_list) and
                     all(isinstance(value, str) for value in event_data[0]) and all(isinstance(value, str) for value in event_data[1])) or (
                            len(event_data) == 1 and isinstance(event_data[0], list) and len(event_data[0]) == len(self.target_path_list)
                            and all(isinstance(value, str) for value in event_data[0])) or len(event_data) == 0)):
            msg = 'Event_data has the wrong format.' \
                'The supported formats are [], [path_value_list] and [path_value_list_1, path_value_list_2], ' \
                'where the path value lists are lists of strings with the same length as the defined paths in the config.'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        # Convert path value lists to tuples
        for i in range(len(event_data)):
            event_data[i] = tuple(event_data[i])

        if len(event_data) == 0:
            # Print the set of all appeared path values if no event_data is given
            values_set = set(self.time_matrix.keys())
            for value in list(values_set):
                for value_2 in self.time_matrix[value]:
                    values_set.add(value_2)

            values_list = list(values_set)
            values_list.sort()

            string = f'Persistence includes transition times to the following path values: {values_list}'
        elif len(event_data) == 1:
            # Print the set of all path values which have a transition time to the path value specified in event_data
            # Check if the path value has an entry in self.time_matrix
            if event_data[0] in self.time_matrix:
                values_set = set(self.time_matrix[event_data[0]].keys())
            else:
                values_set = set()

            # Check if key values in self.time_matrix contain the path value of event_data
            for value in list(self.time_matrix.keys()):
                if event_data[0] in self.time_matrix[value]:
                    values_set.add(value)

            values_list = list(values_set)
            values_list.sort()

            # Set output string
            if len(values_set) > 0:
                string = f'Persistence includes transition times from {event_data[0]} to the following path values: {values_list}'
            else:
                string = f'Persistence includes no transition time from {event_data[0]}.'
        else:
            # Print the transition time
            # Check in which order the event_values appear in the time matrix
            event_value_1 = None
            event_value_2 = None

            if event_data[0] in self.time_matrix and event_data[1] in self.time_matrix[event_data[0]]:
                event_value_1 = event_data[0]
                event_value_2 = event_data[1]
            elif event_data[1] in self.time_matrix and event_data[0] in self.time_matrix[event_data[1]]:
                event_value_1 = event_data[1]
                event_value_2 = event_data[0]

            # Set output string
            if event_value_1 is None:
                string = f'No transition time for {list(event_data[0])} - {list(event_data[1])}.'
            else:
                string = f'Transition time {list(event_data[0])} - {list(event_data[1])}: {self.time_matrix[event_value_1][event_value_2]}.'
        return string

    def add_to_persistence_event(self, event_type, event_data):
        """
        Add or overwrite the information of event_data to the persistence of component_name.
        @return a message with information about the addition to the persistence.
        @throws Exception when the addition of this special event using given event_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        # Query if event_data has the stated format
        if not (isinstance(event_data, list) and len(event_data) == 3 and isinstance(event_data[0], list) and
                isinstance(event_data[1], list) and len(event_data[0]) == len(self.target_path_list) and
                len(event_data[1]) == len(self.target_path_list) and all(isinstance(value, str) for value in event_data[0]) and
                all(isinstance(value, str) for value in event_data[1]) and isinstance(event_data[2], (int, float))):
            msg = 'Event_data has the wrong format.' \
                'The supported format is [path_value_list_1, path_value_list_2, new_transition_time], ' \
                'where the path value lists are lists of strings with the same length as the defined paths in the config.'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        # Convert path value lists to tuples
        event_data[0] = tuple(event_data[0])
        event_data[1] = tuple(event_data[1])

        # Check in which order the event_values appear in the time matrix
        event_value_1 = None
        event_value_2 = None

        if event_data[0] in self.time_matrix and event_data[1] in self.time_matrix[event_data[0]]:
            event_value_1 = event_data[0]
            event_value_2 = event_data[1]
        elif event_data[1] in self.time_matrix and event_data[0] in self.time_matrix[event_data[1]]:
            event_value_1 = event_data[1]
            event_value_2 = event_data[0]

        if event_value_1 is None:
            # Initialize the entry in the time matrix
            if event_data[0] not in self.time_matrix:
                self.time_matrix[event_data[0]] = {}

            self.time_matrix[event_data[0]][event_data[1]] = float(event_data[2])

            return f'Added transition time: {list(event_data[0])} - {list(event_data[1])}, {float(event_data[2])}'

        old_transition_time = self.time_matrix[event_value_1][event_value_2]
        self.time_matrix[event_value_1][event_value_2] = float(event_data[2])
        return f'Changed transition time {list(event_data[0])} - {list(event_data[1])} from {old_transition_time} to {float(event_data[2])}'

    def remove_from_persistence_event(self, event_type, event_data):
        """
        Removes the information of event_data from the persistence of component_name.
        @return a message with information about the removal from the persistence.
        @throws Exception when the addition of this special event using given event_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        # Query if event_data has the stated format
        if not (len(event_data) == 2 and isinstance(event_data[0], list) and isinstance(event_data[1], list) and
                len(event_data[0]) == len(self.target_path_list) and len(event_data[1]) == len(self.target_path_list) and
                all(isinstance(value, str) for value in event_data[0]) and all(isinstance(value, str) for value in event_data[1])):
            msg = 'Event_data has the wrong format. ' \
                'The supported format is [path_value_list_1, path_value_list_2], ' \
                'where the path value lists are lists of strings with the same length as the defined paths in the config.'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)

        # Convert path value lists to tuples
        event_data[0] = tuple(event_data[0])
        event_data[1] = tuple(event_data[1])

        # Check in which order the event_values appear in the time matrix
        event_value_1 = None
        event_value_2 = None

        if event_data[0] in self.time_matrix and event_data[1] in self.time_matrix[event_data[0]]:
            event_value_1 = event_data[0]
            event_value_2 = event_data[1]
        elif event_data[1] in self.time_matrix and event_data[0] in self.time_matrix[event_data[1]]:
            event_value_1 = event_data[1]
            event_value_2 = event_data[0]

        # Check if the transition time between the path values exists
        if event_value_1 is None:
            string = f'Transition time for {list(event_data[0])} - {list(event_data[1])} does not exist and therefore could not be deleted.'
        else:
            # Delete the transition time
            deleted_time = self.time_matrix[event_value_1].pop(event_value_2)

            # Delete the entry to event_value_1 if it is empty
            if self.time_matrix[event_value_1] == {}:
                self.time_matrix.pop(event_value_1)

            string = f'Deleted transition time {list(event_data[0])} - {list(event_data[1])}: {deleted_time}.'

        return string

    def blocklist_event(self, event_type, event_data, blocklisting_data):
        """
        Blocklist an event generated by this source using the information emitted when generating the event.
        @return a message with information about blocklisting
        @throws Exception when blocklisting of this special event using given blocklisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if blocklisting_data is not None:
            msg = 'Blocklisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.ignore_list:
            self.ignore_list.append(event_data)
        return f'Blocklisted path {event_data}.'

    def print(self, message, log_atom, affected_path, confidence=None, additional_information=None):
        """Print the message."""
        if isinstance(affected_path, str):
            affected_path = [affected_path]
        if additional_information is None:
            additional_information = {}

        original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
        if original_log_line_prefix is None:
            original_log_line_prefix = ''

        if self.output_logline:
            tmp_str = ''
            for x in list(log_atom.parser_match.get_match_dictionary().keys()):
                tmp_str += '  ' + x + os.linesep
            tmp_str = tmp_str.lstrip('  ')
            sorted_log_lines = [tmp_str + original_log_line_prefix + log_atom.raw_data.decode()]
            analysis_component = {'AffectedLogAtomPaths': list(log_atom.parser_match.get_match_dictionary().keys())}
        else:
            tmp_str = ''
            for x in affected_path:
                tmp_str += '  ' + x + os.linesep
            tmp_str = tmp_str.lstrip('  ')
            sorted_log_lines = [tmp_str + log_atom.raw_data.decode()]
            analysis_component = {'AffectedLogAtomPaths': affected_path}

        for key, value in additional_information.items():
            analysis_component[key] = value

        event_data = {'AnalysisComponent': analysis_component, 'TypeInfo': {}}
        if confidence is not None:
            event_data['TypeInfo']['Confidence'] = confidence
        for listener in self.anomaly_event_handlers:
            listener.receive_event(f'Analysis.{self.__class__.__name__}', message, sorted_log_lines, event_data, log_atom, self)



"""
This module defines a debug model element that can be used to check whether a specific position in the parsing tree is reached by log atoms.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import logging
from aminer.AminerConfig import DEBUG_LOG_NAME
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ModelElementInterface import ModelElementInterface


class DebugModelElement(ModelElementInterface):
    """
    This class defines a model element matching any data of length zero at any position.
    Thus, it can never fail to match and can be inserted at any position in the parsing tree, where matching itself does not alter parsing
    flow (see e.g. FirstMatchModelElement). It will immediately write the current state of the match to stderr for inspection.
    """

    def __init__(self, element_id: str):
        """
        Initialize the ModelElement.
        @param element_id an identifier for the ModelElement which is shown in the path.
        """
        super().__init__(element_id)
        # To avoid having those elements hidden in production configuration, write a line every time the class is instantiated.
        msg = f"DebugModelElement {element_id} added"
        logging.getLogger(DEBUG_LOG_NAME).info(msg)
        print(msg, file=sys.stderr)

    def get_match_element(self, path: str, match_context):
        """
        @param path to be printed in the MatchElement.
        @param match_context the match_context to be analyzed.
        @return Always return a match.
        """
        msg = f'DebugModelElement path = "{path}/{self.element_id}", unmatched = "{repr(match_context.match_data)}"'
        logging.getLogger(DEBUG_LOG_NAME).info(msg)
        print(msg, file=sys.stderr)
        return MatchElement(f"{path}/{self.element_id}", b"", b"", None)



import unittest
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase, DummyMatchContext


class FixedWordlistDataModelElementTest(TestBase):
    """Unittests for the FixedWordlistDataModelElement."""

    id_ = "wordlist"
    path = "path"
    wordlist = [b"wordlist", b"word"]

    def test1get_match_element_valid_match(self):
        """Parse matching substring from MatchContext and check if the MatchContext was updated with all characters."""
        data = b"wordlist, word"
        index = 0
        value = b"wordlist"
        match_context = DummyMatchContext(data)
        fixed_wordlist_dme = FixedWordlistDataModelElement(self.id_, self.wordlist)
        match_element = fixed_wordlist_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, index, None)

        data = b"word, wordlist"
        index = 1
        value = b"word"
        match_context = DummyMatchContext(data)
        fixed_wordlist_dme = FixedWordlistDataModelElement(self.id_, self.wordlist)
        match_element = fixed_wordlist_dme.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, index, None)

    def test2get_match_element_no_match(self):
        """Parse not matching substring from MatchContext and check if the MatchContext was not changed."""
        data = b"string wordlist"
        match_context = DummyMatchContext(data)
        fixed_wordlist_dme = FixedWordlistDataModelElement(self.id_, self.wordlist)
        match_element = fixed_wordlist_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"wor wordlist"
        match_context = DummyMatchContext(data)
        fixed_wordlist_dme = FixedWordlistDataModelElement(self.id_, self.wordlist)
        match_element = fixed_wordlist_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"0 wordlist"
        match_context = DummyMatchContext(data)
        fixed_wordlist_dme = FixedWordlistDataModelElement(self.id_, self.wordlist)
        match_element = fixed_wordlist_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"1 word"
        match_context = DummyMatchContext(data)
        fixed_wordlist_dme = FixedWordlistDataModelElement(self.id_, self.wordlist)
        match_element = fixed_wordlist_dme.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test3element_id_input_validation(self):
        """Check if element_id is validated."""
        self.assertRaises(ValueError, FixedWordlistDataModelElement, "", self.wordlist)  # empty element_id
        self.assertRaises(TypeError, FixedWordlistDataModelElement, None, self.wordlist)  # None element_id
        self.assertRaises(TypeError, FixedWordlistDataModelElement, b"path", self.wordlist)  # bytes element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, True, self.wordlist)  # boolean element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, 123, self.wordlist)  # integer element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, 123.22, self.wordlist)  # float element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, {"id": "path"}, self.wordlist)  # dict element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, ["path"], self.wordlist)  # list element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, [], self.wordlist)  # empty list element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, (), self.wordlist)  # empty tuple element_id is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, set(), self.wordlist)  # empty set element_id is not allowed

    def test4wordlist_input_validation(self):
        """Check if wordlist is validated."""
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, "path")  # string wordlist
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, None)  # None wordlist
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, b"path")  # bytes wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, True)  # boolean wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, 123)  # integer wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, 123.22)  # float wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, {"id": "path"})  # dict wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, ["path", "path2"])  # list wordlist with strings not allowed
        self.assertRaises(ValueError, FixedWordlistDataModelElement, self.id_, [b"word", b"path", b"path-like"])  # wrong word order
        self.assertRaises(ValueError, FixedWordlistDataModelElement, self.id_, [b"wordlist", b"word", b"word dictionary"])  # wrong order
        self.assertRaises(ValueError, FixedWordlistDataModelElement, self.id_, [])  # empty list wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, ())  # empty tuple wordlist is not allowed
        self.assertRaises(TypeError, FixedWordlistDataModelElement, self.id_, set())  # empty set wordlist is not allowed

    def test5get_match_element_match_context_input_validation(self):
        """Check if an exception is raised, when other classes than MatchContext are used in get_match_element."""
        model_element = FixedWordlistDataModelElement(self.id_, self.wordlist)
        data = b"abcdefghijklmnopqrstuvwxyz.!?"
        model_element.get_match_element(self.path, DummyMatchContext(data))
        model_element.get_match_element(self.path, MatchContext(data))

        self.assertRaises(AttributeError, model_element.get_match_element, self.path, MatchElement(None, data, None, None))
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data.decode())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123.22)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, True)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, None)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, [])
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, {"key": MatchContext(data)})
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, set())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, ())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, model_element)


if __name__ == "__main__":
    unittest.main()



"""
This module defines an evaluator and generator for event rules.
The overall idea of generation is
1) For each processed event A, randomly select another event B occurring within queue_delta_time.
2) If B chronologically occurs after A, create the hypothesis A => B (observing event A implies that event B must be observed within
current_time+queue_delta_time). If B chronologically occurs before A, create the hypothesis B <= A (observing event A implies that event B
must be observed within currentTime-queueDeltaTime).
3) Observe for a long time (max_observations) whether the hypothesis holds.
4) If the hypothesis holds, transform it to a rule. Otherwise, discard the hypothesis.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from collections import deque
import random
import math
import time
import logging

from aminer.AminerConfig import build_persistence_file_name, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD,\
    STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class EventCorrelationDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """This class tries to find time correlation patterns between different log atom events."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, anomaly_event_handlers, target_path_list=None, max_hypotheses=1000, hypothesis_max_delta_time=5.0,
                 generation_probability=1.0, generation_factor=1.0, max_observations=500, p0=0.9, alpha=0.05, candidates_size=10,
                 hypotheses_eval_delta_time=120.0, delta_time_to_discard_hypothesis=180.0, check_rules_flag=False,
                 learn_mode=True, ignore_list=None, persistence_id='Default', output_logline=True, constraint_list=None,
                 stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that all values occurring in these paths are
               considered for value range generation.
        @param max_hypotheses maximum amount of hypotheses and rules hold in memory.
        @param hypothesis_max_delta_time time span of events considered for hypothesis generation.
        @param generation_probability probability in [0, 1] that currently processed log line is considered for hypothesis with each of the
               candidates.
        @param generation_factor likelihood in [0, 1] that currently processed log line is added to the set of candidates for hypothesis
               generation.
        @param max_observations maximum amount of evaluations before hypothesis is transformed into a rule or discarded or rule is
               evaluated.
        @param p0 expected value for hypothesis evaluation distribution.
        @param alpha confidence value for hypothesis evaluation.
        @param candidates_size maximum number of stored candidates used for hypothesis generation.
        @param hypotheses_eval_delta_time duration between hypothesis evaluation phases that remove old hypotheses that are likely to remain
               unused.
        @param delta_time_to_discard_hypothesis time span required for old hypotheses to be discarded.
        @param check_rules_flag specifies whether existing rules are evaluated.
        @param learn_mode specifies whether new hypotheses are generated.
        @param ignore_list list of paths that are not considered for correlation, i.e., events that contain one of these paths are omitted.
               The default value is [] as None is not iterable.
        @param persistence_id name of persistence file.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            mutable_default_args=["target_path_list", "ignore_list", "constraint_list"], aminer_config=aminer_config,
            anomaly_event_handlers=anomaly_event_handlers, target_path_list=target_path_list, max_hypotheses=max_hypotheses,
            hypothesis_max_delta_time=hypothesis_max_delta_time, generation_probability=generation_probability,
            generation_factor=generation_factor, max_observations=max_observations, p0=p0, alpha=alpha, candidates_size=candidates_size,
            hypotheses_eval_delta_time=hypotheses_eval_delta_time, delta_time_to_discard_hypothesis=delta_time_to_discard_hypothesis,
            check_rules_flag=check_rules_flag, learn_mode=learn_mode, ignore_list=ignore_list, persistence_id=persistence_id,
            output_logline=output_logline, constraint_list=constraint_list, stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        self.last_unhandled_match = None
        self.total_records = 0
        self.forward_hypotheses = {}
        self.back_hypotheses = {}
        self.forward_hypotheses_inv = {}
        self.back_hypotheses_inv = {}
        self.last_hypotheses_eval_timestamp = -1.0
        self.forward_rule_queue = deque([])
        self.back_rule_queue = deque([])
        self.forward_hypotheses_queue = deque([])
        self.back_hypotheses_queue = deque([])
        self.hypothesis_candidates = deque([])
        self.sum_unstable_unknown_hypotheses = 0
        self.last_event_occurrence = {}
        self.min_eval_true_dict = {}
        self.min_eval_true_dict_max_size = 1000
        self.sample_events = {}
        self.back_rules = {}
        self.forward_rules = {}
        self.back_rules_inv = {}
        self.forward_rules_inv = {}

        # Compute the initial minimum amount of positive evaluations for hypotheses to become rules.
        # For rules, this value can be different and will be computed based on the sample observations.
        self.min_eval_true = self.get_min_eval_true(self.max_observations, self.p0, self.alpha)
        self.log_forward_rules_learned = 0
        self.log_back_rules_learned = 0
        self.log_new_forward_rules = []
        self.log_new_back_rules = []

        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        PersistenceUtil.add_persistable_component(self)
        self.persistence_id = persistence_id
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)

        if persistence_data is not None:
            for record in persistence_data:
                implication_direction = record[0]
                trigger_event = tuple(record[1])
                implied_event = tuple(record[2])
                max_obs = record[3]
                min_eval_t = record[4]
                rule = Implication(trigger_event, implied_event, None, max_obs, min_eval_t)
                rule.stable = 1
                if implication_direction == 'back':
                    if trigger_event in self.back_rules:
                        self.back_rules[trigger_event].append(rule)
                    else:
                        self.back_rules[trigger_event] = [rule]
                    if implied_event in self.back_rules_inv:
                        self.back_rules_inv[implied_event].append(rule)
                    else:
                        self.back_rules_inv[implied_event] = [rule]
                elif implication_direction == 'forward':
                    if trigger_event in self.forward_rules:
                        self.forward_rules[trigger_event].append(rule)
                    else:
                        self.forward_rules[trigger_event] = [rule]
                    if implied_event in self.forward_rules_inv:
                        self.forward_rules_inv[implied_event].append(rule)
                    else:
                        self.forward_rules_inv[implied_event] = [rule]
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')

    # skipcq: PYL-R1710
    def get_min_eval_true(self, max_observations, p0, alpha):
        """
        Compute the critical value (minimal amount of true evaluations) for a hypothesis.
        The form of the hypothesis is <eventA> implies <eventB> with at least probability p0 to be accepted.
        This method tries to be efficient by
        - Storing already computed critical values in a dictionary
        - Swapping (1 - p0) and p0 and replace alpha with (1 - alpha) to reduce loops
        """
        if (max_observations, p0, alpha) in self.min_eval_true_dict:
            return self.min_eval_true_dict[(max_observations, p0, alpha)]

        sum1 = 0.0
        max_observations_factorial = math.factorial(max_observations)
        i_factorial = 1
        for i in range(max_observations + 1):
            i_factorial = i_factorial * max(i, 1)
            # No float conversion possible for huge numbers; use integer division.
            sum1 = sum1 + max_observations_factorial / (i_factorial * math.factorial(max_observations - i)) * ((1 - p0) ** i) * (
                      p0 ** (max_observations - i))
            if sum1 > (1 - alpha):
                if len(self.min_eval_true_dict) <= self.min_eval_true_dict_max_size:
                    # Store common values for fast retrieval
                    self.min_eval_true_dict[(max_observations, p0, alpha)] = max_observations - i
                return max_observations - i

    def receive_atom(self, log_atom):
        """Receive a log atom from a source."""
        self.log_total += 1
        if log_atom.get_timestamp() is None:
            log_atom.atom_time = time.time()

        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False

        parser_match = log_atom.parser_match
        self.total_records += 1

        # Skip paths from ignore_list.
        for ignore_path in self.ignore_list:
            if ignore_path in parser_match.get_match_dictionary():
                return
        if self.target_path_list is None or len(self.target_path_list) == 0:
            # Event is defined by the full path of log atom.
            constraint_path_flag = False
            for constraint_path in self.constraint_list:
                if parser_match.get_match_dictionary().get(constraint_path) is not None:
                    constraint_path_flag = True
                    break
            if not constraint_path_flag and self.constraint_list != []:
                return
            log_event = tuple(parser_match.get_match_dictionary())
        else:
            # Event is defined by value combos in target_path_list
            values = []
            all_values_none = True
            for path in self.target_path_list:
                match = parser_match.get_match_dictionary().get(path)
                if match is None:
                    continue
                matches = []
                if isinstance(match, list):
                    matches = match
                else:
                    matches.append(match)
                for match in matches:
                    if isinstance(match.match_object, bytes):
                        value = match.match_object.decode(AminerConfig.ENCODING)
                    else:
                        value = str(match.match_object)
                    if value is not None:
                        all_values_none = False
                    values.append(value)
            if all_values_none is True:
                return
            log_event = tuple(values)

        # Store last seen sample event to improve output.
        self.sample_events[log_event] = log_atom.raw_data

        if self.check_rules_flag:
            # Only check rules without generating new hypotheses.

            # Trigger implication A => B when A occurs.
            if log_event in self.forward_rules:
                for rule in self.forward_rules[log_event]:
                    rule.rule_trigger_timestamps.append(log_atom.atom_time)
                    self.forward_rule_queue.append(rule)

            # Resolve triggered implication A => B when B occurs.
            if log_event in self.forward_rules_inv:
                for rule in self.forward_rules_inv[log_event]:
                    # Find first non-observed trigger timestamp
                    trigger_timestamp_index = -1
                    for trigger_timestamp in rule.rule_trigger_timestamps:
                        trigger_timestamp_index += 1
                        if trigger_timestamp != 'obs':
                            break
                    if trigger_timestamp_index != -1 and \
                            rule.rule_trigger_timestamps[trigger_timestamp_index] != 'obs' and \
                            rule.rule_trigger_timestamps[trigger_timestamp_index] >= log_atom.atom_time - self.hypothesis_max_delta_time:
                        # Implication was triggered; append positive evaluation and mark as seen.
                        rule.add_rule_observation(1)
                        rule.rule_trigger_timestamps[trigger_timestamp_index] = 'obs'

            # Clean up triggered/resolved implications.
            while len(self.forward_rule_queue) > 0:
                rule = self.forward_rule_queue[0]
                if len(rule.rule_trigger_timestamps) == 0:
                    # Triggered timestamp was already deleted somewhere else.
                    self.forward_rule_queue.popleft()
                    continue
                if rule.rule_trigger_timestamps[0] == 'obs':
                    # Remove triggered timestamp.
                    rule.rule_trigger_timestamps.popleft()
                    self.forward_rule_queue.popleft()
                    continue
                if rule.rule_trigger_timestamps[0] < log_atom.atom_time - self.hypothesis_max_delta_time:
                    # Too much time has elapsed; append negative evaluation.
                    rule.add_rule_observation(0)
                    rule.rule_trigger_timestamps.popleft()
                    self.forward_rule_queue.popleft()
                    if not rule.evaluate_rule():
                        try:
                            data = log_atom.raw_data.decode(AminerConfig.ENCODING)
                        except UnicodeError:
                            data = repr(log_atom.raw_data)
                        original_log_line_prefix = self.aminer_config.config_properties.get(
                            CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
                        tmp_string = f'Rule: {str(rule.trigger_event)} -> {str(rule.implied_event)}\n  Expected: ' \
                                     f'{str(rule.min_eval_true)}/{str(rule.max_observations)}\n  Observed: ' \
                                     f'{str(sum(rule.rule_observations))}/{str(len(rule.rule_observations))}'
                        if self.output_logline:
                            sorted_log_lines = [tmp_string + '\n' + original_log_line_prefix + data]
                        else:
                            sorted_log_lines = [tmp_string + data]
                        for listener in self.anomaly_event_handlers:
                            implied_event = None
                            trigger_event = None
                            if rule.implied_event in self.sample_events:
                                implied_event = self.sample_events[rule.implied_event]
                            if rule.trigger_event in self.sample_events:
                                trigger_event = self.sample_events[rule.trigger_event]
                            listener.receive_event(
                                'analysis.EventCorrelationDetector',
                                f'Correlation rule violated! Event {repr(implied_event)} is missing, but should follow event '
                                f'{repr(trigger_event)}', sorted_log_lines,
                                {'RuleInfo': {'Rule': str(rule.trigger_event) + '->' + str(rule.implied_event),
                                              'Expected': str(rule.min_eval_true) + '/' + str(rule.max_observations),
                                              'Observed': str(sum(rule.rule_observations)) + '/' + str(len(rule.rule_observations))}},
                                log_atom, self)
                        rule.rule_observations = deque([])
                    continue
                break

            # Trigger implication B <= A when B occurs.
            if log_event in self.back_rules_inv:
                for rule in self.back_rules_inv[log_event]:
                    rule.rule_trigger_timestamps.append(log_atom.atom_time)
                    self.back_rule_queue.append(rule)

            # Resolve triggered implication B <= A when A occurs.
            if log_event in self.back_rules:
                for rule in self.back_rules[log_event]:
                    # Find first non-observed trigger timestamp
                    trigger_timestamp_index = -1
                    for trigger_timestamp in rule.rule_trigger_timestamps:
                        trigger_timestamp_index += 1
                        if trigger_timestamp != 'obs':
                            break
                    if trigger_timestamp_index != -1 and \
                            rule.rule_trigger_timestamps[trigger_timestamp_index] != 'obs' and \
                            rule.rule_trigger_timestamps[trigger_timestamp_index] >= log_atom.atom_time - self.hypothesis_max_delta_time:
                        rule.add_rule_observation(1)
                        rule.rule_trigger_timestamps[trigger_timestamp_index] = 'obs'
                    else:
                        rule.add_rule_observation(0)
                        if not rule.evaluate_rule():
                            try:
                                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
                            except UnicodeError:
                                data = repr(log_atom.raw_data)
                            original_log_line_prefix = self.aminer_config.config_properties.get(
                                CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
                            tmp_string = f'Rule: {str(rule.implied_event)} <- {str(rule.trigger_event)}\n  Expected: ' \
                                         f'{str(rule.min_eval_true)}/{str(rule.max_observations)}\n  Observed: ' \
                                         f'{str(sum(rule.rule_observations))}/{str(len(rule.rule_observations))}'
                            if self.output_logline:
                                sorted_log_lines = [tmp_string + '\n' + original_log_line_prefix + data]
                            else:
                                sorted_log_lines = [tmp_string + data]
                            for listener in self.anomaly_event_handlers:
                                implied_event = None
                                trigger_event = None
                                if rule.implied_event in self.sample_events:
                                    implied_event = self.sample_events[rule.implied_event]
                                if rule.trigger_event in self.sample_events:
                                    trigger_event = self.sample_events[rule.trigger_event]
                                listener.receive_event(
                                    'analysis.EventCorrelationDetector',
                                    f'Correlation rule violated! Event {repr(implied_event)} is missing, but should precede event '
                                    f'{repr(trigger_event)}', sorted_log_lines,
                                    {'RuleInfo': {'Rule': str(rule.implied_event) + '<-' + str(rule.trigger_event),
                                                  'Expected': str(rule.min_eval_true) + '/' + str(rule.max_observations),
                                                  'Observed': str(sum(rule.rule_observations)) + '/' + str(len(rule.rule_observations))}},
                                    log_atom, self)
                            rule.rule_observations = deque([])

            # Clean up triggered/resolved implications.
            while len(self.back_rule_queue) > 0:
                rule = self.back_rule_queue[0]
                if len(rule.rule_trigger_timestamps) == 0:
                    self.back_rule_queue.popleft()
                    continue
                if rule.rule_trigger_timestamps[0] == 'obs':
                    rule.rule_trigger_timestamps.popleft()
                    self.back_rule_queue.popleft()
                    continue
                if rule.rule_trigger_timestamps[0] < log_atom.atom_time - self.hypothesis_max_delta_time:
                    rule.rule_trigger_timestamps.popleft()
                    self.back_rule_queue.popleft()
                    continue
                break

        if self.learn_mode:
            # Generate new hypotheses and rules.

            # Keep track of event occurrences, relevant for removing old hypotheses.
            self.last_event_occurrence[log_event] = log_atom.atom_time

            # Trigger implication A => B when A occurs.
            if log_event in self.forward_hypotheses:
                for implication in self.forward_hypotheses[log_event]:
                    if implication.stable == 0:
                        implication.hypothesis_trigger_timestamps.append(log_atom.atom_time)
                        self.forward_hypotheses_queue.append(implication)

            # Resolve triggered implication A => B when B occurs.
            if log_event in self.forward_hypotheses_inv:
                delete_hypotheses = []
                for implication in self.forward_hypotheses_inv[log_event]:
                    # Find first non-observed trigger timestamp
                    trigger_timestamp_index = -1
                    for trigger_timestamp in implication.hypothesis_trigger_timestamps:
                        trigger_timestamp_index += 1
                        if trigger_timestamp != 'obs':
                            break
                    if trigger_timestamp_index != -1 and \
                            str(implication.hypothesis_trigger_timestamps[trigger_timestamp_index]) != 'obs' and \
                            implication.hypothesis_trigger_timestamps[trigger_timestamp_index] >= log_atom.atom_time - \
                            self.hypothesis_max_delta_time and \
                            implication.stable == 0:
                        implication.add_hypothesis_observation(1, log_atom.atom_time)
                        # Mark this timestamp as observed
                        implication.hypothesis_trigger_timestamps[trigger_timestamp_index] = 'obs'
                        # Since only true observations occur here, check for instability not necessary.
                        if implication.compute_hypothesis_stability() == 1:
                            # Update p and min_eval_true according to the results in the sample.
                            p = implication.hypothesis_evaluated_true / implication.hypothesis_observations
                            implication.min_eval_true = self.get_min_eval_true(self.max_observations, p, self.alpha)
                            # Add hypothesis to rules.
                            if implication.trigger_event in self.forward_rules:
                                self.forward_rules[implication.trigger_event].append(implication)
                                self.log_forward_rules_learned += 1
                                self.log_new_forward_rules.append(implication)
                            else:
                                self.forward_rules[implication.trigger_event] = [implication]
                                self.log_forward_rules_learned += 1
                                self.log_new_forward_rules.append(implication)
                            if implication.implied_event in self.forward_rules_inv:
                                self.forward_rules_inv[implication.implied_event].append(implication)
                            else:
                                self.forward_rules_inv[implication.implied_event] = [implication]
                            # Drop time stamps of previous observations, start new observations for rule.
                            implication.hypothesis_trigger_timestamps.clear()
                            self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses - 1
                            # Remove implication from list of hypotheses.
                            self.forward_hypotheses[implication.trigger_event].remove(implication)
                            delete_hypotheses.append(implication)
                for delete_hypothesis in delete_hypotheses:
                    self.forward_hypotheses_inv[log_event].remove(delete_hypothesis)

            # Clean up triggered/resolved implications.
            while len(self.forward_hypotheses_queue) > 0:
                implication = self.forward_hypotheses_queue[0]
                if len(implication.hypothesis_trigger_timestamps) == 0:
                    # Triggered timestamp was already deleted somewhere else.
                    self.forward_hypotheses_queue.popleft()
                    continue
                if implication.hypothesis_trigger_timestamps[0] == 'obs':
                    # Remove triggered timestamp.
                    implication.hypothesis_trigger_timestamps.popleft()
                    self.forward_hypotheses_queue.popleft()
                    continue
                if implication.hypothesis_trigger_timestamps[0] < log_atom.atom_time - self.hypothesis_max_delta_time:
                    # Too much time has elapsed; append negative evaluation.
                    implication.hypothesis_trigger_timestamps.popleft()
                    implication.add_hypothesis_observation(0, log_atom.atom_time)
                    if implication.compute_hypothesis_stability() == -1 and implication.trigger_event in self.forward_hypotheses and \
                            implication in self.forward_hypotheses[implication.trigger_event]:
                        # This check is required if a hypothesis was already removed, but triggered hypotheses are still in the queue.
                        self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses - 1
                        self.forward_hypotheses[implication.trigger_event].remove(implication)
                        self.forward_hypotheses_inv[implication.implied_event].remove(implication)
                        if len(self.forward_hypotheses[implication.trigger_event]) == 0:
                            del self.forward_hypotheses[implication.trigger_event]
                        if len(self.forward_hypotheses_inv[implication.implied_event]) == 0:
                            del self.forward_hypotheses_inv[implication.implied_event]
                    self.forward_hypotheses_queue.popleft()
                    continue
                break

            # Trigger implication B <= A when B occurs.
            if log_event in self.back_hypotheses_inv:
                for implication in self.back_hypotheses_inv[log_event]:
                    if implication.stable == 0:
                        implication.hypothesis_trigger_timestamps.append(log_atom.atom_time)
                        self.back_hypotheses_queue.append(implication)

            # Resolve triggered implication B <= A when A occurs.
            if log_event in self.back_hypotheses:
                delete_hypotheses = []
                for implication in self.back_hypotheses[log_event]:
                    if implication.stable == 0:
                        # Find first non-observed trigger timestamp
                        trigger_timestamp_index = -1
                        for trigger_timestamp in implication.hypothesis_trigger_timestamps:
                            trigger_timestamp_index += 1
                            if trigger_timestamp != 'obs':
                                break
                        if trigger_timestamp_index != -1 and \
                                str(implication.hypothesis_trigger_timestamps[trigger_timestamp_index]) != 'obs' and \
                                implication.hypothesis_trigger_timestamps[trigger_timestamp_index] >= log_atom.atom_time - \
                                self.hypothesis_max_delta_time:
                            implication.add_hypothesis_observation(1, log_atom.atom_time)
                            implication.hypothesis_trigger_timestamps[trigger_timestamp_index] = 'obs'
                            # Since only true observations occur here, check for instability not necessary.
                            if implication.compute_hypothesis_stability() == 1:
                                # Update p and min_eval_true according to the results in the sample.
                                p = implication.hypothesis_evaluated_true / implication.hypothesis_observations
                                implication.min_eval_true = self.get_min_eval_true(self.max_observations, p, self.alpha)
                                # Add hypothesis to rules.
                                if implication.trigger_event in self.back_rules:
                                    self.back_rules[implication.trigger_event].append(implication)
                                    self.log_back_rules_learned += 1
                                    self.log_new_back_rules.append(implication)
                                else:
                                    self.back_rules[implication.trigger_event] = [implication]
                                    self.log_back_rules_learned += 1
                                    self.log_new_back_rules.append(implication)
                                if implication.implied_event in self.back_rules_inv:
                                    self.back_rules_inv[implication.implied_event].append(implication)
                                else:
                                    self.back_rules_inv[implication.implied_event] = [implication]
                                # Drop time stamps of previous observations, start new observations for rule.
                                implication.hypothesis_trigger_timestamps.clear()
                                self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses - 1
                                # Remove implication from list of hypotheses.
                                delete_hypotheses.append(implication)
                                self.back_hypotheses_inv[implication.implied_event].remove(implication)
                        else:
                            implication.add_hypothesis_observation(0, log_atom.atom_time)
                            if implication.compute_hypothesis_stability() == -1:
                                self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses - 1
                                delete_hypotheses.append(implication)
                                self.back_hypotheses_inv[implication.implied_event].remove(implication)
                                if len(self.back_hypotheses_inv[implication.implied_event]) == 0:
                                    del self.back_hypotheses_inv[implication.implied_event]
                for delete_hypothesis in delete_hypotheses:
                    self.back_hypotheses[log_event].remove(delete_hypothesis)
                if len(self.back_hypotheses[log_event]) == 0:
                    del self.back_hypotheses[log_event]

            # Clean up triggered/resolved implications.
            while len(self.back_hypotheses_queue) > 0:
                implication = self.back_hypotheses_queue[0]
                if len(implication.hypothesis_trigger_timestamps) == 0:
                    self.back_hypotheses_queue.popleft()
                    continue
                if implication.hypothesis_trigger_timestamps[0] == 'obs':
                    implication.hypothesis_trigger_timestamps.popleft()
                    self.back_hypotheses_queue.popleft()
                    continue
                if implication.hypothesis_trigger_timestamps[0] < log_atom.atom_time - self.hypothesis_max_delta_time:
                    implication.hypothesis_trigger_timestamps.popleft()
                    self.back_hypotheses_queue.popleft()
                    continue
                break

            # Generate new hypotheses
            if len(self.hypothesis_candidates) > 0 and random.uniform(0.0, 1.0) < self.generation_factor:
                implication_direction = random.randint(0, 1)
                if self.sum_unstable_unknown_hypotheses >= self.max_hypotheses:
                    # If too many hypotheses exist, do nothing.
                    implication_direction = -1
                if implication_direction == 0:
                    for candidate in self.hypothesis_candidates:
                        candidate_event = candidate[0]
                        # Chronological implication is: candidate_event <= log_event
                        implication = Implication(log_event, candidate_event, log_atom.atom_time, self.max_observations, self.min_eval_true)
                        if log_event in self.back_hypotheses:
                            # Only add hypotheses that are not already present as hypotheses.
                            continue_outer = False
                            for imp in self.back_hypotheses[log_event]:
                                if candidate_event == imp.implied_event:
                                    continue_outer = True
                                    break
                            if continue_outer:
                                continue
                        if log_event in self.back_rules:
                            # Only add hypotheses that are not already present as rules.
                            continue_outer = False
                            for imp in self.back_rules[log_event]:
                                if candidate_event == imp.implied_event:
                                    continue_outer = True
                                    break
                            if continue_outer:
                                continue
                        # At this point it is known that the implication is new, otherwise a continue statement would have been reached
                        if log_event in self.back_hypotheses:
                            self.back_hypotheses[log_event].append(implication)
                        else:
                            self.back_hypotheses[log_event] = [implication]
                        if candidate_event in self.back_hypotheses_inv:
                            self.back_hypotheses_inv[candidate_event].append(implication)
                        else:
                            self.back_hypotheses_inv[candidate_event] = [implication]
                        self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses + 1
                elif implication_direction == 1:
                    for candidate in self.hypothesis_candidates:
                        candidate_event = candidate[0]
                        # Chronological implication is: candidate_event => log_event
                        # Skip event A => event A since already covered by back hypotheses
                        if log_event != candidate_event:
                            implication = Implication(candidate_event, log_event, log_atom.atom_time, self.max_observations,
                                                      self.min_eval_true)
                            if candidate_event in self.forward_hypotheses:
                                # Only add hypotheses that are not already present as hypotheses.
                                continue_outer = False
                                for imp in self.forward_hypotheses[candidate_event]:
                                    if log_event == imp.implied_event:
                                        continue_outer = True
                                        break
                                if continue_outer:
                                    continue
                            if candidate_event in self.forward_rules:
                                # Only add hypotheses that are not already present as rules.
                                continue_outer = False
                                for imp in self.forward_rules[candidate_event]:
                                    if log_event == imp.implied_event:
                                        continue_outer = True
                                        break
                                if continue_outer:
                                    continue
                            # At this point it is known that the implication is new, otherwise a continue statement would have been reached
                            if candidate_event in self.forward_hypotheses:
                                self.forward_hypotheses[candidate_event].append(implication)
                            else:
                                self.forward_hypotheses[candidate_event] = [implication]
                            if log_event in self.forward_hypotheses_inv:
                                self.forward_hypotheses_inv[log_event].append(implication)
                            else:
                                self.forward_hypotheses_inv[log_event] = [implication]
                            self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses + 1
                if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                    self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time

            # Periodically remove old or unstable hypotheses.
            if log_atom.atom_time >= self.last_hypotheses_eval_timestamp + self.hypotheses_eval_delta_time:
                self.last_hypotheses_eval_timestamp = log_atom.atom_time
                empty_back_events = []
                for event in self.back_hypotheses:
                    outdated_hypotheses_indexes = []
                    i = 0
                    for implication in self.back_hypotheses[event]:
                        if implication.stable == 0 and self.last_event_occurrence[
                                event] < log_atom.atom_time - self.delta_time_to_discard_hypothesis:
                            self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses - 1
                            outdated_hypotheses_indexes.append(i)
                            self.back_hypotheses_inv[implication.implied_event].remove(implication)
                            if len(self.back_hypotheses_inv[implication.implied_event]) == 0:
                                del self.back_hypotheses_inv[implication.implied_event]
                        i = i + 1
                    # Reverse list to avoid index changes after deletions.
                    for outdated_hypothesis_index in reversed(outdated_hypotheses_indexes):
                        del self.back_hypotheses[event][outdated_hypothesis_index]
                    if len(self.back_hypotheses[event]) == 0:
                        empty_back_events.append(event)
                for empty_back_event in empty_back_events:
                    del self.back_hypotheses[empty_back_event]
                empty_forward_events = []
                for event in self.forward_hypotheses:
                    outdated_hypotheses_indexes = []
                    i = 0
                    for implication in self.forward_hypotheses[event]:
                        if implication.stable == 0 and implication.most_recent_observation_timestamp < log_atom.atom_time -\
                                self.delta_time_to_discard_hypothesis:
                            self.sum_unstable_unknown_hypotheses = self.sum_unstable_unknown_hypotheses - 1
                            outdated_hypotheses_indexes.append(i)
                            self.forward_hypotheses_inv[implication.implied_event].remove(implication)
                            if len(self.forward_hypotheses_inv[implication.implied_event]) == 0:
                                del self.forward_hypotheses_inv[implication.implied_event]
                        i = i + 1
                    # Reverse list to avoid index changes after deletions.
                    for outdated_hypothesis_index in reversed(outdated_hypotheses_indexes):
                        del self.forward_hypotheses[event][outdated_hypothesis_index]
                    if len(self.forward_hypotheses[event]) == 0:
                        empty_forward_events.append(event)
                for empty_forward_event in empty_forward_events:
                    del self.forward_hypotheses[empty_forward_event]

            # Remove old hypothesis candidates
            while len(self.hypothesis_candidates) > 0:
                candidate = self.hypothesis_candidates[0]
                if candidate[1] < log_atom.atom_time - self.hypothesis_max_delta_time:
                    self.hypothesis_candidates.popleft()
                    continue
                break

            # Add new hypothesis candidates
            if len(self.hypothesis_candidates) < self.candidates_size and random.uniform(0.0, 1.0) < self.generation_probability:
                self.hypothesis_candidates.append((log_event, log_atom.atom_time))
        self.log_success += 1

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        known_path_set = set()
        for event_a in self.back_rules:
            for implication in self.back_rules[event_a]:
                known_path_set.add(
                    ('back', tuple(event_a), tuple(implication.implied_event), implication.max_observations, implication.min_eval_true))
        for event_a in self.forward_rules:
            for implication in self.forward_rules[event_a]:
                known_path_set.add(
                    ('forward', tuple(event_a), tuple(implication.implied_event), implication.max_observations, implication.min_eval_true))
        PersistenceUtil.store_json(self.persistence_file_name, list(known_path_set))
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_forward_rules_learned} new forward rules and {self.log_back_rules_learned} new back rules in the last 60 "
                f"minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and"
                f" learned {self.log_forward_rules_learned} new forward rules and {self.log_back_rules_learned} new back rules in the last "
                f"60 minutes. Following new forward rules were learned: {self.log_forward_rules_learned}. Following new back rules were"
                f" learned: {self.log_back_rules_learned}")
        self.log_success = 0
        self.log_total = 0
        self.log_forward_rules_learned = 0
        self.log_back_rules_learned = 0
        self.log_new_forward_rules = []
        self.log_new_back_rules = []

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.constraint_list:
            self.constraint_list.append(event_data)
        return f'Allowlisted path {event_data}.'

    def blocklist_event(self, event_type, event_data, blocklisting_data):
        """
        Blocklist an event generated by this source using the information emitted when generating the event.
        @return a message with information about blocklisting
        @throws Exception when blocklisting of this special event using given blocklisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if blocklisting_data is not None:
            msg = 'Blocklisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.ignore_list:
            self.ignore_list.append(event_data)
        return f'Blocklisted path {event_data}.'


class Implication:
    """Define the shape of an implication rule."""

    def __init__(self, trigger_event, implied_event, generation_time, max_observations, min_eval_true):
        self.trigger_event = trigger_event
        self.implied_event = implied_event
        self.stable = 0  # 0 .. unknown, 1 .. stable, -1 .. unstable
        self.max_observations = max_observations
        self.min_eval_true = min_eval_true
        self.most_recent_observation_timestamp = generation_time
        self.hypothesis_trigger_timestamps = deque([])
        self.rule_trigger_timestamps = deque([])
        self.rule_observations = deque([])
        # Hypothesis is only generated for observed implication. Thus, initialized with 1.
        self.hypothesis_observations = 1
        self.hypothesis_evaluated_true = 1

    def add_hypothesis_observation(self, result, timestamp):
        """Update the observation counts for a hypothesis."""
        # Reset counters when max_observations is reached.
        self.most_recent_observation_timestamp = timestamp
        if self.hypothesis_observations >= self.max_observations:
            pass
        else:
            self.hypothesis_observations = self.hypothesis_observations + 1
            self.hypothesis_evaluated_true = self.hypothesis_evaluated_true + result

    def compute_hypothesis_stability(self):
        """Compute the stability of a hypothesis."""
        if self.hypothesis_evaluated_true >= self.min_eval_true:
            # Known that hypothesis is stable.
            self.stable = 1
        elif (self.hypothesis_observations - self.hypothesis_evaluated_true) > (self.max_observations - self.min_eval_true):
            # Known that hypothesis will never be stable.
            self.stable = -1
        else:
            # Stability is still unknown, more observations required.
            self.stable = 0
        return self.stable

    def add_rule_observation(self, result):
        """Add a new rule to the observations."""
        if len(self.rule_observations) >= self.max_observations:
            self.rule_observations.popleft()
        self.rule_observations.append(result)

    def evaluate_rule(self):
        """Evaluate a rule."""
        ones = 0
        for obs in self.rule_observations:
            ones = ones + obs
        return (len(self.rule_observations) - ones) <= (self.max_observations - self.min_eval_true)

    def __repr__(self):
        return str(self.trigger_event[-1]).split('/')[-1] + '->' + str(self.implied_event[-1]).split('/')[-1] + ', eval=' + str(
            self.hypothesis_evaluated_true) + '/' + str(self.hypothesis_observations) + ', rule=' + str(
            self.rule_observations) + ', ruletriggerts=' + str(self.rule_trigger_timestamps)

    def get_dictionary_repr(self):
        """Return the dictionary representation of an Implication."""
        return {'trigger_event': self.trigger_event, 'implied_event': self.implied_event, 'stable': self.stable,
                'max_observations': self.max_observations, 'min_eval_true': self.min_eval_true,
                'most_recent_observation_timestamp': self.most_recent_observation_timestamp,
                'hypothesis_trigger_timestamps': list(self.hypothesis_trigger_timestamps),
                'rule_trigger_timestamps': list(self.rule_trigger_timestamps), 'rule_observations': list(self.rule_observations),
                'hypothesis_observations': self.hypothesis_observations, 'hypothesis_evaluated_true': self.hypothesis_evaluated_true}


def set_random_seed(seed):
    """Set the random seed for testing purposes."""
    random.seed(seed)



"""
This module defines a detector for new values in a data path.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import logging

from aminer.AminerConfig import build_persistence_file_name, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD,\
    STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class NewMatchPathValueDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """This class creates events when new values for a given data path were found."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, target_path_list, anomaly_event_handlers, persistence_id='Default', learn_mode=False,
                 output_logline=True, stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that all values occurring in these paths are
               considered for value range generation.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param persistence_id name of persistence file.
        @param learn_mode when set to True, this detector will report a new value only the first time before including it
        in the known values set automatically.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            aminer_config=aminer_config, target_path_list=target_path_list, anomaly_event_handlers=anomaly_event_handlers,
            persistence_id=persistence_id, learn_mode=learn_mode, output_logline=output_logline, stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        self.log_learned_path_values = 0
        self.log_new_learned_values = []

        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        PersistenceUtil.add_persistable_component(self)
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        if persistence_data is None:
            self.known_values_set = set()
        else:
            self.known_values_set = set(persistence_data)
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')

    def receive_atom(self, log_atom):
        """Receive a log atom from a source."""
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False

        for target_path in self.target_path_list:
            match = match_dict.get(target_path)
            if match is None:
                continue
            matches = []
            if isinstance(match, list):
                matches = match
            else:
                matches.append(match)
            affected_log_atom_values = []
            for match in matches:
                if match.match_object not in self.known_values_set:
                    if self.learn_mode:
                        self.known_values_set.add(match.match_object)
                        self.log_learned_path_values += 1
                        self.log_new_learned_values.append(match.match_object)
                        if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                            self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time

                    if isinstance(match.match_object, bytes):
                        affected_log_atom_values.append(match.match_object.decode(AminerConfig.ENCODING))
                    else:
                        affected_log_atom_values.append(str(match.match_object))
            if len(affected_log_atom_values) > 0:
                analysis_component = {'AffectedLogAtomPaths': [target_path], 'AffectedLogAtomValues': affected_log_atom_values}
                if isinstance(match_dict.get(target_path), list):
                    res = {target_path: affected_log_atom_values}
                else:
                    res = {target_path: match_dict.get(target_path).match_object}
                    if isinstance(res[target_path], bytes):
                        res[target_path] = res[target_path].decode(AminerConfig.ENCODING)
                try:
                    data = log_atom.raw_data.decode(AminerConfig.ENCODING)
                except UnicodeError:
                    data = repr(log_atom.raw_data)
                if self.output_logline:
                    original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
                    sorted_log_lines = [str(res) + os.linesep + original_log_line_prefix + data]
                else:
                    sorted_log_lines = [str(res)]
                event_data = {'AnalysisComponent': analysis_component}
                for listener in self.anomaly_event_handlers:
                    listener.receive_event(f'Analysis.{self.__class__.__name__}', 'New value(s) detected', sorted_log_lines, event_data,
                                           log_atom, self)
                self.log_success += 1

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        PersistenceUtil.store_json(self.persistence_file_name, list(self.known_values_set))
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        self.known_values_set.add(event_data)
        return f"Allowlisted path(es) {', '.join(self.target_path_list)} with {event_data}."

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_values} new values in the last 60 minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_values} new values in the last 60 minutes. Following new value combinations were"
                f" learned: {self.log_new_learned_values}")
        self.log_success = 0
        self.log_total = 0
        self.log_learned_path_values = 0
        self.log_new_learned_values = []



"""
This component counts occurring combinations of values and periodically sends the results as a report.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import logging
from aminer.AminerConfig import DEBUG_LOG_NAME
from aminer.AnalysisChild import AnalysisContext
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


current_processed_lines_str = 'CurrentProcessedLines'
total_processed_lines_str = 'TotalProcessedLines'


class ParserCount(AtomHandlerInterface, TimeTriggeredComponentInterface):
    """This class creates a counter for path value combinations."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, target_path_list, anomaly_event_handlers, report_interval=60, target_label_list=None,
                 split_reports_flag=False):
        """
        Initialize the ParserCount component.
        @param aminer_config configuration from analysis_context.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that all values occurring in these paths are
               considered for value range generation.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param report_interval delay in seconds before reporting.
        @param target_label_list a list of labels for the target_path_list. This list must have the same size as target_path_list.
        @param split_reports_flag if true every path produces an own report, otherwise one report for all paths is produced.
        """
        # avoid "defined outside init" issue
        self.log_success, self.log_total = [None]*2
        super().__init__(
            mutable_default_args=["target_path_list"], aminer_config=aminer_config, target_path_list=target_path_list,
            anomaly_event_handlers=anomaly_event_handlers, report_interval=report_interval, target_label_list=target_label_list,
            split_reports_flag=split_reports_flag
        )
        self.count_dict = {}
        self.next_report_time = None
        if (self.target_path_list is None or self.target_path_list == []) and (
                self.target_label_list is not None and self.target_label_list != []):
            msg = 'Target labels cannot be used without specifying target paths.'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)
        if self.target_label_list is not None and len(self.target_path_list) != len(self.target_label_list):
            msg = 'Every path must have a target label if target labels are used.'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise ValueError(msg)

        for target_path in self.target_path_list:
            if self.target_label_list:
                target_path = self.target_label_list[self.target_path_list.index(target_path)]
            self.count_dict[target_path] = {current_processed_lines_str: 0, total_processed_lines_str: 0}

    def receive_atom(self, log_atom):
        """Receive a log atom from a source."""
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        success_flag = False
        for target_path in self.target_path_list:
            match_element = match_dict.get(target_path)
            if match_element is not None:
                success_flag = True
                if self.target_label_list:
                    target_path = self.target_label_list[self.target_path_list.index(target_path)]
                self.count_dict[target_path][current_processed_lines_str] += 1
                self.count_dict[target_path][total_processed_lines_str] += 1
        if not self.target_path_list:
            path = iter(match_dict).__next__()
            if path not in self.count_dict:
                self.count_dict[path] = {current_processed_lines_str: 0, total_processed_lines_str: 0}
            self.count_dict[path][current_processed_lines_str] += 1
            self.count_dict[path][total_processed_lines_str] += 1

        if self.next_report_time is None:
            self.next_report_time = time.time() + self.report_interval
        if success_flag:
            self.log_success += 1
        return True

    def do_timer(self, trigger_time):
        """Check current ruleset should be persisted."""
        if self.next_report_time is None:
            return self.report_interval

        delta = self.next_report_time - trigger_time
        if delta <= 0:
            self.send_report()
            delta = self.report_interval
            self.next_report_time = time.time() + delta
        return delta

    # skipcq: PYL-R0201
    def do_persist(self):
        """Immediately write persistence data to storage."""
        return False

    def send_report(self):
        """Send a report to the event handlers."""
        output_string = 'Parsed paths in the last ' + str(self.report_interval) + ' seconds:\n'
        if not self.split_reports_flag:
            for k in self.count_dict:
                c = self.count_dict[k]
                output_string += '\t' + str(k) + ': ' + str(c) + '\n'
            output_string = output_string[:-1]
            event_data = {'StatusInfo': self.count_dict, 'FromTime': time.time() - self.report_interval, 'ToTime': time.time()}
            for listener in self.anomaly_event_handlers:
                listener.receive_event(f'Analysis.{self.__class__.__name__}', 'Count report', [output_string], event_data, None, self)
        else:
            for k in self.count_dict:
                output_string = 'Parsed paths in the last ' + str(self.report_interval) + ' seconds:\n'
                c = self.count_dict[k]
                output_string += '\t' + str(k) + ': ' + str(c)
                status_info = {k: {
                    current_processed_lines_str: c[current_processed_lines_str],
                    total_processed_lines_str: c[total_processed_lines_str]}}
                event_data = {'StatusInfo': status_info, 'FromTime': time.time() - self.report_interval, 'ToTime': time.time()}
                for listener in self.anomaly_event_handlers:
                    listener.receive_event(f'Analysis.{self.__class__.__name__}', 'Count report', [output_string], event_data, None, self)
        for k in self.count_dict:
            self.count_dict[k][current_processed_lines_str] = 0
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} sent report.')



"""
This module defines an detector for event and value sequences.
The concept is based on STIDE which was first published by Forrest, S.,
Hofmeyr, S. A., Somayaji, A., & Longstaff, T. A. (1996, May). A sense of self
for unix processes. In Proceedings of the 1996 IEEE Symposium on Security and
Privacy (pp. 120-128). IEEE.
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import logging

from aminer.AminerConfig import build_persistence_file_name, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD,\
    STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class EventSequenceDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """This class creates events when new event or value sequences were found."""

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, anomaly_event_handlers, id_path_list=None, target_path_list=None, seq_len=3, allow_missing_id=False,
                 timeout=-1, persistence_id='Default', learn_mode=False, output_logline=True, ignore_list=None,
                 constraint_list=None, stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param target_path_list parser paths of values to be analyzed. Multiple paths mean that values are analyzed by their combined
               occurrences. When no paths are specified, the events given by the full path list are analyzed.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param id_path_list one or more paths that specify the trace of the sequence detection, i.e., incorrect sequences that
               are generated by interleaved events can be avoided when event sequence identifiers are available.
        @param seq_len the length of the sequences to be learned (larger lengths increase precision, but may overfit the data).
        @param allow_missing_id specifies whether log atoms without id path should be omitted (only if id path is set).
        @param timeout maximum allowed seconds between two entries of sequence; sequence is split in subsequences if exceeded.
        @param persistence_id name of persistence file.
        @param learn_mode specifies whether new frequency measurements override ground truth frequencies.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param ignore_list list of paths that are not considered for analysis, i.e., events that contain one of these paths are omitted.
               The default value is [] as None is not iterable.
        @param constraint_list list of paths that have to be present in the log atom to be analyzed.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            mutable_default_args=["id_path_list", "target_path_list", "ignore_list", "constraint_list"], aminer_config=aminer_config,
            anomaly_event_handlers=anomaly_event_handlers, id_path_list=id_path_list, target_path_list=target_path_list, seq_len=seq_len,
            allow_missing_id=allow_missing_id, timeout=timeout, persistence_id=persistence_id, learn_mode=learn_mode,
            output_logline=output_logline, ignore_list=ignore_list, constraint_list=constraint_list,
            stop_learning_time=stop_learning_time, stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        self.sequences = set()
        self.current_sequences = {}
        self.last_seen_times = {}
        self.log_learned = 0
        self.log_learned_sequences = []

        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        PersistenceUtil.add_persistable_component(self)

        # Persisted data contains lists of sequences, i.e., [[<seq1_elem1>, <seq1_elem2>], [<seq2_elem1, ...], ...]
        # Thereby, sequence elements may be tuples, i.e., combinations of values, or paths that define events.
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        if persistence_data is not None:
            for sequence in persistence_data:
                sequence_elem_tuple = []
                for sequence_elem in sequence:
                    sequence_elem_tuple.append(tuple(sequence_elem))
                self.sequences.add(tuple(sequence_elem_tuple))
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')

    def receive_atom(self, log_atom):
        """Receive a log atom from a source."""
        parser_match = log_atom.parser_match
        self.log_total += 1
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False

        # Skip paths from ignore list.
        for ignore_path in self.ignore_list:
            if ignore_path in parser_match.get_match_dictionary().keys():
                return

        if self.target_path_list is None or len(self.target_path_list) == 0:
            # Event is defined by the full path of log atom.
            constraint_path_flag = False
            for constraint_path in self.constraint_list:
                if parser_match.get_match_dictionary().get(constraint_path) is not None:
                    constraint_path_flag = True
                    break
            if not constraint_path_flag and self.constraint_list != []:
                return
            log_event = tuple(parser_match.get_match_dictionary().keys())
        else:
            # Event is defined by value combos in target_path_list
            values = []
            all_values_none = True
            for path in self.target_path_list:
                match = parser_match.get_match_dictionary().get(path)
                if match is None:
                    continue
                matches = []
                if isinstance(match, list):
                    matches = match
                else:
                    matches.append(match)
                # the match variable is not needed any more and reused for the iteration.
                for match in matches:
                    if isinstance(match.match_object, bytes):
                        value = match.match_object.decode(AminerConfig.ENCODING)
                    else:
                        value = str(match.match_object)
                    if value is not None:
                        all_values_none = False
                    values.append(value)
            if all_values_none is True:
                return
            log_event = tuple(values)

        # In case that id_path_list is set, use it to differentiate sequences by their id.
        # Otherwise, the empty tuple () is used as the only key of the current_sequences dict.
        id_tuple = ()
        for id_path in self.id_path_list:
            id_match = parser_match.get_match_dictionary().get(id_path)
            if id_match is None:
                if self.allow_missing_id is True:
                    # Insert placeholder for id_path that is not available
                    id_tuple += ('',)
                else:
                    # Omit log atom if one of the id paths is not found.
                    return
            else:
                matches = []
                if isinstance(id_match, list):
                    matches = id_match
                else:
                    matches.append(id_match)
                for match in matches:
                    if isinstance(match.match_object, bytes):
                        id_tuple += (match.match_object.decode(AminerConfig.ENCODING),)
                    else:
                        id_tuple += (match.match_object,)

        # Create entry for the id_tuple in the current_sequences dict if it did not occur before.
        if id_tuple not in self.current_sequences:
            self.current_sequences[id_tuple] = ()

        # If too much time passed between two values, start a new sequence
        if self.timeout != -1:
            if id_tuple in self.last_seen_times and self.last_seen_times[id_tuple] is not None and \
                    log_atom.atom_time is not None and self.last_seen_times[id_tuple] + self.timeout < log_atom.atom_time:
                self.current_sequences[id_tuple] = ()
            self.last_seen_times[id_tuple] = log_atom.atom_time

        # If the sequence has not reached its full length, append the newest element and stop.
        # Otherwise, the current sequence is used as a queue, where the oldest entry is removed.
        if len(self.current_sequences[id_tuple]) < self.seq_len:
            self.current_sequences[id_tuple] += (log_event,)
            if len(self.current_sequences[id_tuple]) != self.seq_len:
                return
        else:
            self.current_sequences[id_tuple] = self.current_sequences[id_tuple][1:] + (log_event,)

        # Report anomalies if the current processed sequence never occurred before.
        if self.current_sequences[id_tuple] not in self.sequences:
            if self.learn_mode is True:
                self.sequences.add(self.current_sequences[id_tuple])
                self.log_learned += 1
                self.log_learned_sequences.append(self.current_sequences[id_tuple])
                if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                    self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
            if self.output_logline:
                sorted_log_lines = [log_atom.parser_match.match_element.annotate_match('') + os.linesep + original_log_line_prefix +
                                    data]
            else:
                sorted_log_lines = [data]
            if self.target_path_list is None or len(self.target_path_list) == 0:
                analysis_component = {'AffectedLogAtomPaths': self.current_sequences[id_tuple]}
            else:
                analysis_component = {'AffectedLogAtomPaths': self.target_path_list,
                                      'AffectedLogAtomValues': list(self.current_sequences[id_tuple])}
            if self.id_path_list is not None:
                analysis_component['AffectedIdValues'] = list(id_tuple)
            event_data = {'AnalysisComponent': analysis_component}
            for listener in self.anomaly_event_handlers:
                listener.receive_event(f'Analysis.{self.__class__.__name__}', 'New sequence detected', sorted_log_lines, event_data,
                                       log_atom, self)
        self.log_success += 1

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        PersistenceUtil.store_json(self.persistence_file_name, list(self.sequences))
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.constraint_list:
            self.constraint_list.append(event_data)
        return f'Allowlisted path {event_data}.'

    def blocklist_event(self, event_type, event_data, blocklisting_data):
        """
        Blocklist an event generated by this source using the information emitted when generating the event.
        @return a message with information about blocklisting
        @throws Exception when blocklisting of this special event using given blocklisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if blocklisting_data is not None:
            msg = 'Blocklisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.ignore_list:
            self.ignore_list.append(event_data)
        return f'Blocklisted path {event_data}.'

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned} new sequences in the last 60 minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned} new sequences in the last 60 minutes. Following new sequences were learned: "
                f"{str(self.log_learned_sequences)}")
        self.log_success = 0
        self.log_total = 0
        self.log_learned = 0
        self.log_learned_sequences = []



"""
This file defines the NewMatchIdValueComboDetector.
detector to extract values from multiple LogAtoms and check, if the value
combination was already seen before.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import logging

from aminer.AminerConfig import build_persistence_file_name, DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD,\
    STAT_LOG_NAME, CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig
from aminer.AnalysisChild import AnalysisContext
from aminer.events.EventInterfaces import EventSourceInterface
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer.util.TimeTriggeredComponentInterface import TimeTriggeredComponentInterface


class NewMatchIdValueComboDetector(AtomHandlerInterface, TimeTriggeredComponentInterface, EventSourceInterface):
    """
    This class creates events when a new value combination for a given list of match data.
    Paths need to be found in log atoms with the same id value in a specific path.
    """

    time_trigger_class = AnalysisContext.TIME_TRIGGER_CLASS_REALTIME

    def __init__(self, aminer_config, target_path_list, anomaly_event_handlers, id_path_list, min_allowed_time_diff,
                 persistence_id='Default', allow_missing_values_flag=False, learn_mode=False, output_logline=True,
                 stop_learning_time=None, stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param aminer_config configuration from analysis_context.
        @param target_path_list the list of values to extract from each match to create the value combination to be checked.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        @param id_path_list the list of paths where id values can be stored in all relevant log event types.
        @param min_allowed_time_diff the minimum amount of time in seconds after the first appearance of a log atom with a specific id
               that is waited for other log atoms with the same id to occur. The maximum possible time to keep an incomplete combo
               is 2*min_allowed_time_diff
        @param persistence_id name of persistence file.
        @param allow_missing_values_flag when set to True, the detector will also use matches, where one of the paths from target_path_list
               does not refer to an existing parsed data object.
        @param learn_mode when set to True, this detector will report a new value only the first time before including it
               in the known values set automatically.
        @param output_logline specifies whether the full parsed log atom should be provided in the output.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        super().__init__(
            aminer_config=aminer_config, target_path_list=target_path_list, anomaly_event_handlers=anomaly_event_handlers,
            id_path_list=id_path_list, min_allowed_time_diff=min_allowed_time_diff, persistence_id=persistence_id,
            allow_missing_values_flag=allow_missing_values_flag, learn_mode=learn_mode, output_logline=output_logline,
            stop_learning_time=stop_learning_time, stop_learning_no_anomaly_time=stop_learning_no_anomaly_time
        )
        self.log_learned_path_value_combos = 0
        self.log_new_learned_values = []

        self.persistence_file_name = build_persistence_file_name(aminer_config, self.__class__.__name__, persistence_id)
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        self.known_values = []
        if persistence_data is not None:
            # Combinations are stored as list of dictionaries
            for record in persistence_data:
                self.known_values.append(record)
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')
        PersistenceUtil.add_persistable_component(self)

        self.id_dict_current = {}
        self.id_dict_old = {}
        self.next_shift_time = None

    def receive_atom(self, log_atom):
        """
        Receive on parsed atom and the information about the parser match.
        @return True if a value combination was extracted and checked against the list of known combinations, no matter if the checked
                values were new or not.
        """
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False

        id_match_element = None
        for id_path in self.id_path_list:
            # Get the id value and return if not found in this log atom.
            id_match_element = match_dict.get(id_path)
            if id_match_element is not None:
                break
        if id_match_element is None:
            return False

        timestamp = log_atom.get_timestamp()
        if timestamp is not None:
            if self.next_shift_time is None:
                self.next_shift_time = timestamp + self.min_allowed_time_diff
            if timestamp > self.next_shift_time:
                # Every min_allowed_time_diff seconds, process all combinations from id_dict_old and then override id_dict_old with
                # id_dict_current. This guarantees that incomplete combos are hold for at least min_allowed_time_diff seconds before
                # proceeding.
                self.next_shift_time = timestamp + self.min_allowed_time_diff
                if self.allow_missing_values_flag:
                    for id_old in self.id_dict_old:
                        self.process_id_dict_entry(self.id_dict_old[id_old], log_atom)
                self.id_dict_old = self.id_dict_current
                self.id_dict_current = {}

        if isinstance(id_match_element, list):
            id_match_object = []
            for match_element in id_match_element:
                id_match_object.append(match_element.match_object)
            id_match_object = tuple(id_match_object)
        else:
            id_match_object = id_match_element.match_object

        # Find dictionary containing id and create ref to old or current dict (side effects)
        id_dict = None
        if id_match_object in self.id_dict_current:
            id_dict = self.id_dict_current
        elif id_match_object in self.id_dict_old:
            id_dict = self.id_dict_old
        else:
            id_dict = self.id_dict_current
            id_dict[id_match_object] = {}

        for target_path in self.target_path_list:
            # Append values to the combo.
            match_element = match_dict.get(target_path)
            if match_element is not None:
                if isinstance(match_element, list):
                    values = []
                    matches = match_element
                    for match_element in matches:
                        if isinstance(match_element.match_object, bytes):
                            values.append(match_element.match_object.decode(AminerConfig.ENCODING))
                        else:
                            values.append(id_dict[id_match_object][target_path])
                    id_dict[id_match_object][target_path] = values
                else:
                    if isinstance(match_element.match_object, bytes):
                        id_dict[id_match_object][target_path] = match_element.match_object.decode(AminerConfig.ENCODING)
                    else:
                        id_dict[id_match_object][target_path] = match_element.match_object

        if len(id_dict[id_match_object]) == len(self.target_path_list):
            # Found value for all target paths. No need to wait more.
            self.process_id_dict_entry(id_dict[id_match_object], log_atom)
            del id_dict[id_match_object]
        self.log_success += 1
        return True

    def process_id_dict_entry(self, id_dict_entry, log_atom):
        """Process an entry from the id_dict."""
        if id_dict_entry not in self.known_values:
            # Combo is unknown, process and raise anomaly
            if self.learn_mode:
                self.known_values.append(id_dict_entry)
                self.log_learned_path_value_combos += 1
                self.log_new_learned_values.append(id_dict_entry)
                if self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                    self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time

            analysis_component = {'AffectedLogAtomValues': [str(i) for i in list(id_dict_entry.values())]}
            event_data = {'AnalysisComponent': analysis_component}
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
            if self.output_logline:
                sorted_log_lines = [log_atom.parser_match.match_element.annotate_match('') + os.linesep + repr(
                    id_dict_entry) + os.linesep + original_log_line_prefix + data]
            else:
                sorted_log_lines = [repr(id_dict_entry) + os.linesep + original_log_line_prefix + data]
            for listener in self.anomaly_event_handlers:
                listener.receive_event(f'Analysis.{self.__class__.__name__}', 'New value combination(s) detected', sorted_log_lines,
                                       event_data, log_atom, self)

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        PersistenceUtil.store_json(self.persistence_file_name, self.known_values)
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if event_data not in self.known_values:
            self.known_values.append(event_data)
        return f"Allowlisted path(es) {', '.join(self.target_path_list)} with {event_data}."

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_value_combos} new value combinations in the last 60 minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_value_combos} new value combinations in the last 60 minutes. Following new value combinations "
                f"were learned: {self.log_new_learned_values}")
        self.log_success = 0
        self.log_total = 0
        self.log_learned_path_value_combos = 0
        self.log_new_learned_values = []
